# (c) 2013 Marcos Dione <mdione@grulic.org.ar>

# This file is part of ayrton.
#
# ayrton is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ayrton is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ayrton.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import io
from collections.abc import Iterable
import signal
import errno
from traceback import format_exc

import ayrton

import logging
logger= logging.getLogger ('ayrton.execute')

encoding= sys.getdefaultencoding ()

# special value to signal that the output should be captured
# instead of going to stdout
class Capture:
    pass

# same, but for a pipe
class Pipe:
    pass

class o (object):
    def __init__ (self, **kwargs):
        option= list (kwargs.items ())[0]
        self.key=   option[0]
        self.value= option[1]

class CommandFailed (Exception):
    def __init__ (self, command):
        self.command= command

    def __str__ (self):
        return "%s: %d" % (' '.join ([ repr (arg) for arg in self.command.args]), self.command._exit_code)

class CommandNotFound (NameError):
    def __init__ (self, name):
        self.name= name

    def __str__ (self):
        return "CommandNotFound or NameError: command %(name)s not found or name %(name)s is not defined" % self.__dict__

def which(program):
    def is_exe(fpath):
        return os.path.exists(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program

    else:
        if "PATH" not in os.environ:
            return None

        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None

def resolve_program(program):
    path = which(program)
    if not path:
        # our actual command might have a dash in it, but we can't call
        # that from python (we have to use underscores), so we'll check
        # if a dash version of our underscore command exists and use that
        # if it does
        if "_" in program:
            path = which(program.replace("_", "-"))

        if not path:
            return None

    return path

class Command:
    default_options= dict (
        _in_tty= False,
        _out_tty= False,
        _env= {},
        _end= os.linesep.encode (encoding),
        _chomp= True,
        _encoding= encoding,
        _bg= False,
        _fails= False,
        )

    supported_options= ('_in', '_out', '_err', '_end', '_chomp', '_encoding',
                        '_env', '_bg', '_fails')

    def __init__ (self, path):
        self.path= path
        self.exe= resolve_program (path)
        self.command= None

        self.stdin_pipe= None
        self.stdout_pipe= None
        self.stderr_pipe= None

        self._exit_code= None
        self.capture_file= None

        self.child_pid= None

    def prepare_fds (self):
        if '_in' in self.options:
            i= self.options['_in']
            logger.debug ('_in: %r', i)

            # this condition is complex, but basically handles any type of input
            # that is not going to be handled later in client()
            if (     not isinstance (i, io.IOBase)      # file and such objects
                 and type (i) not in (int, str, bytes)  # fd's and file names
                 and i is not None                      # special value for /dev/null
                 and not isinstance (i, Command) ):     # Command's
                if self.options['_in_tty']:
                    # TODO: no support yet for input from file when _in_tty
                    # NOTE: os.openpty() returns (master, slave)
                    # but when writing master=w, slave=r
                    logger.debug ('_in: using tty')
                    (master, slave)= os.openpty ()
                    self.stdin_pipe= (slave, master)
                else:
                    logger.debug ("_in:%s creates a pipe()", type (i))
                    self.stdin_pipe= os.pipe ()
            elif isinstance (i, Command):
                if i.options.get ('_out', None)==Capture:
                    # if it's a captured command, create a pipe to feed the data
                    logger.debug ("_in::Command, _in._out==Capture creates a pipe()")
                    self.stdin_pipe= os.pipe ()
                elif i.options.get ('_out', None)==Pipe:
                    # if it's a piped command, use its pipe and hope it runs in the bg
                    logger.debug ("_in::Command uses the stdout_pipe")
                    self.stdin_pipe= i.stdout_pipe

        logger.debug ("stdin_pipe: %s", self.stdin_pipe)

        if '_out' in self.options:
            if self.options['_out']==Capture:
                if self.options['_out_tty']:
                    if self.options['_in_tty']:
                        # we use a copy of the pty created for the stdin
                        logger.debug ('duping tty form _in to _out')
                        self.stdout_pipe= (os.dup (self.stdin_pipe[1]),
                                           os.dup (self.stdin_pipe[0]))
                    else:
                        # this time the order is right
                        logger.debug ('_out: using tty')
                        (master, slave)= os.openpty ()
                        self.stdout_pipe= (master, slave)
                else:
                    logger.debug ("_out==Capture creates a pipe()")
                    self.stdout_pipe= os.pipe ()
            elif self.options['_out']==Pipe:
                # this pipe should be picked up by the outer Command
                logger.debug ("_out==Pipe creates a pipe()")
                self.stdout_pipe= os.pipe ()

        logger.debug ("stdout_pipe: %s", self.stdout_pipe)

        if '_err' in self.options:
            if self.options['_err']==Capture:
                # if stdout is also Capture'd, then use the same pipe
                if not '_out' in self.options or self.options['_out']!=Capture:
                    # if stdout is a tty, hook to that one
                    if self.options['_out_tty']:
                        logger.debug ('duping tty form _out to _err')
                        self.stderr_pipe= (os.dup (self.stdout_pipe[0]),
                                           os.dup (self.stdout_pipe[1]))
                    else:
                        logger.debug ("_err==Capture creates a pipe()")
                        self.stderr_pipe= os.pipe ()
            elif self.options['_err']==Pipe:
                # this pipe should be picked up by the outer Command
                logger.debug ("_err==Pipe creates a pipe()")
                self.stderr_pipe= os.pipe ()

        logger.debug ("stderr_pipe: %s", self.stderr_pipe)

    def child (self):
        logger.debug ('child')

        ifile= ofile= efile= None  # these hold the IOBase object to close

        try:
            if '_in' in self.options:
                i= self.options['_in']

                if i is None:
                    # connect to /dev/null
                    # it's not /dev/zero, see man (4) zero
                    logger.debug ("_in==None redirects from /dev/null")
                    i= os.open (os.devnull, os.O_RDONLY)
                elif isinstance (i, io.IOBase):
                    # this does not work with file like objects
                    ifile= i
                    logger.debug ("_in::IOBase redirects %s -> 0 (stdin)", ifile)
                    i= i.fileno ()
                elif type (i) in (str, bytes):
                    file_name= i
                    i= os.open (i, os.O_RDONLY)
                    logger.debug ("_in::(str|bytes) redirects %d (%s) -> 0 (stdin)", i, file_name)

                if isinstance (i, int):
                    logger.debug ("_in::int redirects %d -> 0 (stdin)", i)
                    os.dup2 (i, 0)
                    if ifile is None:
                        logger.debug ('closing %d', i)
                        os.close (i)
                    else:
                        # closes both the IOBase file and its fh
                        logger.debug ('closing %s', ifile)
                        ifile.close ()
                else:
                    # use the pipe prepared by prepare_fds()
                    # including the case where _in::Command
                    logger.debug ("_in::%s, redirect stdin_pipe from prepare_fds() to stdin", type (i))
                    r, w= self.stdin_pipe
                    logger.debug ("_in: redirecting %d -> 0", r)
                    os.dup2 (r, 0)
                    logger.debug ('closing %d', r)
                    os.close (r)
                    # the write fd can already be closed in this case:
                    # a= cat (..., _out=capture)
                    # b= grep (..., _in= a)
                    # once a's lines have been read, t
                    os.close (w)

            if '_out' in self.options:
                o= self.options['_out']
                if o is None:
                    # connect to /dev/null
                    o= os.open (os.devnull, os.O_WRONLY) # no need to create it
                    logger.debug ("_out==None, redirects stdout 1 -> %d (%s)", o, os.devnull)
                elif isinstance (o, io.IOBase):
                    # this does not work with file like objects
                    ofile= o
                    logger.debug ("_out::IOBase, redirects stdout 1 -> %s", ofile)
                    o= o.fileno ()
                elif type (o) in (bytes, str):
                    file_name= o
                    o= os.open (o, os.O_WRONLY)
                    logger.debug ("_out::(str|bytes), redirects stdout 1 -> %d (%s)", o, file_name)

                if isinstance (o, int):
                    logger.debug ("_out::int, redirects stdout 1 -> %d", o)
                    os.dup2 (o, 1)
                    if ofile is None:
                        logger.debug ('closing %d', o)
                        os.close (o)
                    else:
                        # closes both the IOBase file and its fh
                        logger.debug ('closing %s', ofile)
                        ofile.close ()
                elif o==Capture or o==Pipe:
                    r, w= self.stdout_pipe
                    logger.debug ("_out::(Capture or Pipe), redirects stdout 1 -> %d", w)
                    os.dup2 (w, 1)
                    logger.debug ('closing %d', w)
                    os.close (w)
                    os.close (r)

            if '_err' in self.options:
                e= self.options['_err']
                if e is None:
                    # connect to /dev/null
                    e= os.open (os.devnull, os.O_WRONLY) # no need to create it
                    logger.debug ("_err==None, redirects stderr 2 -> %d (%s)", e, os.devnull)
                elif isinstance (e, io.IOBase):
                    # this does not work with file like objects
                    efile= e
                    logger.debug ("_err::IOBase, redirects stderr 2 -> %s", efile)
                    e= e.fileno ()
                elif type (e) in (bytes, str):
                    file_name= e
                    e= os.open (e, os.O_WRONLY)
                    logger.debug ("_err::(str|bytes), redirects stderr 2 -> %d (%s)", e, file_name)

                if isinstance (e, int):
                    logger.debug ("_err::int, redirects stderr 2 -> %d", e)
                    os.dup2 (e, 2)
                    if efile is None:
                        logger.debug ('closing %d', e)
                        os.close (e)
                    else:
                        # closes both the IOBase file and its fh
                        logger.debug ('closing %s', efile)
                        efile.close ()
                elif e==Capture:
                    if '_out' in self.options and self.options['_out']==Capture:
                        # send it to stdout
                        logger.debug ("_err::Capture and _out::Capture, redirects stderr 2 -> 1")
                        os.dup2 (1, 2)
                    else:
                        r, w= self.stderr_pipe
                        logger.debug ("_err::Capture, redirects stderr 2 -> %d", w)
                        os.dup2 (w, 2)
                        logger.debug ('closing %d', r)
                        os.close (r)
        except FileNotFoundError as e:
            logger.debug (e)
            # TODO: report something
            os._exit (1)

        # restore some signals
        for i in (signal.SIGPIPE, signal.SIGINT):
            signal.signal (i, signal.SIG_DFL)

        try:
            os.execvpe (self.exe, self.args, self.options['_env'])
        except FileNotFoundError as e:
            logger.debug (e)
            # TODO: report something
            os._exit (127)


    def prepare_args (self, cmd, args, kwargs):
        ans= [cmd]

        for arg in args:
            if isinstance (arg, o):
                self.prepare_arg (ans, arg.key, arg.value)
            else:
                ans.append (str (arg))

        for k, v in kwargs.items ():
            self.prepare_arg (ans, k, v)

        return ans

    def prepare_arg (self, seq, name, value):
        if value!=False:
            seq.append (name)

            # this is not the same as 'not value'
            # because value can have any, well, value of any kind
            if value!=True:
                seq.append (str (value))
        else:
            # TODO: --no-option?
            pass

    def parent (self):
        logger.debug ('parent')

        if self.stdin_pipe is not None:
            # str -> write into the fd
            # list -> write each
            # file -> take fd
            i= self.options['_in']

            r, w= self.stdin_pipe
            logger.debug ('closing %d', r)
            os.close (r)

            if isinstance (i, Iterable) and (not isinstance (i, Command) or
                                             i.options.get ('_out', None)==Capture):
                # this includes file-like's and Capture'd Commands
                for e in i:
                    os.write (w, str (e).encode (encoding))
                    os.write (w, self.options['_end'])
            elif not isinstance (i, Command):
                os.write (w, str (i).encode (encoding))
                os.write (w, self.options['_end'])

            logger.debug ('closing %d', w)
            os.close (w)

        # TODO: The return status of a pipeline is the exit status of the last
        # command, unless the pipefail option is enabled. If pipefail is enabled,
        # the pipeline's return status is the value of the last (rightmost)
        # command to exit with a non-zero status, or zero if all commands exit
        # successfully.
        if not self.options['_bg']:
            self.wait ()
            ayrton.runner.wait_for_pending_children ()
        else:
            ayrton.runner.pending_children.append (self)

    def wait (self):
        self._exit_code= os.waitpid (self.child_pid, 0)[1] >> 8

        reader_pipe= None

        if self.stdout_pipe is not None:
            # this will also read stderr if both are Capture'd
            reader_pipe= self.stdout_pipe
        if self.stderr_pipe is not None:
            reader_pipe= self.stderr_pipe

        if reader_pipe is not None and self.options.get ('_out', None)!=Pipe:
            r, w= reader_pipe
            logger.debug ('closing %d', w)
            os.close (w)
            self.capture_file= open (r)

        if self._exit_code==127:
            # NOTE: when running bash, it returns 127 when it can't find the script to run
            raise CommandNotFound (self.path)

        if (ayrton.runner.options.get ('errexit', False) and
            self._exit_code!=0 and
            not self.options.get ('_fails', False)):

            raise CommandFailed (self)

    def exit_code (self):
        if self._exit_code is None:
            self.wait ()
        return self._exit_code

    def __call__ (self, *args, **kwargs):
        if self.exe is None:
            raise CommandNotFound (self.path)

        self.options= self.default_options.copy ()
        for option in self.supported_options:
            try:
                # update with the passed value
                self.options[option]= kwargs[option]
                # we don't need the option anymore
                del kwargs[option]
            except KeyError:
                # ignore
                pass

        self.options['_env'].update (os.environ)
        self.args= self.prepare_args (self.exe, args, kwargs)

        self.stdin_pipe= None
        self.stdout_pipe= None
        self.stderr_pipe= None

        self._exit_code= None
        self.capture_file= None

        self.prepare_fds ()

        if type (self.options['_end'])!=bytes:
            self.options['_end']= str (self.options['_end']).encode (encoding)

        logger.debug ('fork')
        r= os.fork ()
        if r==0:
            try:
                self.child ()
            except Exception as e:
                logger.debug ('child borked')
                logger.debug (format_exc())

            # catch all
            os._exit (1)
        else:
            self.child_pid= r
            self.parent ()

        return self

    def __bool__ (self):
        if self._exit_code is None:
            self.wait ()
        return self._exit_code==0

    def __str__ (self):
        if self._exit_code is None:
            self.wait ()
        return self.capture_file.read ()

    def __iter__ (self):
        if self._exit_code is None:
            self.wait ()
        if self.capture_file is not None:
            for line in self.capture_file.readlines ():
                if self.options['_chomp']:
                    line= line.rstrip (os.linesep)

                yield line

            # finish him!
            self.capture_file.close ()
        else:
            # TODO
            pass

    # BUG this method is leaking an opend file()
    # self.capture_file
    def readline (self):
        if self._exit_code is None:
            self.wait ()
        line= self.capture_file.readline ()
        if self.options['_chomp']:
            line= line.rstrip (os.linesep)

        return line

    def close (self):
        if self._exit_code is None:
            self.wait ()
        self.capture_file.close ()

    def readlines (self):
        if self._exit_code is None:
            self.wait ()
        # ugly way to not leak the file()
        return ( line for line in self )

    def __del__ (self):
        # finish it
        if self._exit_code is None and self.child_pid is not None:
            self.wait ()



# -*- coding: utf-8 -*-

# (c) 2013 Marcos Dione <mdione@grulic.org.ar>

# This file is part of ayrton.
#
# ayrton is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ayrton is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ayrton.  If not, see <http://www.gnu.org/licenses/>.

import ayrton
import ayrton.execute
import os

import logging
logger= logging.getLogger ('ayrton.functions')

# NOTE: all this code is excuted in the script's environment

class cd (object):
    def __init__ (self, dir):
        self.old_dir= os.getcwd ()
        os.chdir (dir)

    def __enter__ (self):
        pass

    def __exit__ (self, *args):
        os.chdir (self.old_dir)

def export (**kwargs):
    for k, v in kwargs.items ():
        ayrton.runner.globals[k]= str (v)
        os.environ[k]= str (v)

option_map= dict (
    e= 'errexit',
    )

def option (option, value=True):
    if len (option)==2:
        if option[0]=='-':
            value= True
        elif option[0]=='+':
            value= False
        else:
            # TODO: syntax error:
            pass
        option= option_map[option[1]]

    ayrton.runner.options[option]= value

def run (path, *args, **kwargs):
    c= ayrton.execute.Command (path)
    return c (*args, **kwargs)

def shift (n=1):
    # we start at 1 becasuse 0 is the script's path
    # this closely follows bash's behavior
    if n==1:
        ans= ayrton.runner.globals['argv'].pop (1)
    elif n>1:
        ans= [ ayrton.runner.globals['argv'].pop (1)
               for i in range (n) ]
    else:
        raise ValueError ()

    return ans


def unset (*args):
    for k in args:
        if k in ayrton.runner.globals.keys ():
            # found, remove it
            del ayrton.runner.globals[k]
            del os.environ[k]



# ______________________________________________________________________
"""Module pytokenize

THIS FILE WAS COPIED FROM pypy/module/parser/pytokenize.py AND ADAPTED
TO BE ANNOTABLE (Mainly made lists homogeneous)

This is a modified version of Ka-Ping Yee's tokenize module found in the
Python standard library.

The primary modification is the removal of the tokenizer's dependence on the
standard Python regular expression module, which is written in C.  The regular
expressions have been replaced with hand built DFA's using the
basil.util.automata module.

$Id: pytokenize.py,v 1.3 2003/10/03 16:31:53 jriehl Exp $
"""
# ______________________________________________________________________

from ayrton.parser.pyparser import automata
from ayrton.parser.pyparser.dfa_generated import *

__all__ = [ "tokenize" ]

endDFAs = {"'" : singleDFA,
           '"' : doubleDFA,
           'r' : None,
           'R' : None,
           "u" : None,
           "U" : None,
           'b' : None,
           'B' : None}

for uniPrefix in ("", "b", "B"):
    for rawPrefix in ("", "r", "R"):
        prefix_1 = uniPrefix + rawPrefix
        prefix_2 = rawPrefix + uniPrefix

        endDFAs[prefix_1 + "'''"] = single3DFA
        endDFAs[prefix_1 + '"""'] = double3DFA
        endDFAs[prefix_2 + "'''"] = single3DFA
        endDFAs[prefix_2 + '"""'] = double3DFA

for uniPrefix in ("u", "U"):
    endDFAs[uniPrefix + "'''"] = single3DFA
    endDFAs[uniPrefix + '"""'] = double3DFA

whiteSpaceStatesAccepts = [True]
whiteSpaceStates = [{'\t': 0, ' ': 0, '\x0c': 0}]
whiteSpaceDFA = automata.DFA(whiteSpaceStates, whiteSpaceStatesAccepts)

# ______________________________________________________________________
# COPIED:

triple_quoted = {}
for t in ("'''", '"""',
          "r'''", 'r"""', "R'''", 'R"""',
          "u'''", 'u"""', "U'''", 'U"""',
          "b'''", 'b"""', "B'''", 'B"""',
          "br'''", 'br"""', "Br'''", 'Br"""',
          "bR'''", 'bR"""', "BR'''", 'BR"""',
          "rb'''", 'rb"""', "rB'''", 'rB"""',
          "Rb'''", 'Rb"""', "RB'''", 'RB"""'):
    triple_quoted[t] = t
single_quoted = {}
for t in ("'", '"',
          "r'", 'r"', "R'", 'R"',
          "u'", 'u"', "U'", 'U"',
          "b'", 'b"', "B'", 'B"',
          "br'", 'br"', "Br'", 'Br"',
          "bR'", 'bR"', "BR'", 'BR"',
          "rb'", 'rb"', "rB'", 'rB"',
          "Rb'", 'Rb"', "RB'", 'RB"'):

    single_quoted[t] = t

tabsize = 8
alttabsize = 1

# PYPY MODIFICATION: removed TokenError class as it's not needed here

# PYPY MODIFICATION: removed StopTokenizing class as it's not needed here

# PYPY MODIFICATION: removed printtoken() as it's not needed here

# PYPY MODIFICATION: removed tokenize() as it's not needed here

# PYPY MODIFICATION: removed tokenize_loop() as it's not needed here

# PYPY MODIFICATION: removed generate_tokens() as it was copied / modified
#                    in pythonlexer.py

# PYPY MODIFICATION: removed main() as it's not needed here

# ______________________________________________________________________
# End of pytokenize.py




from ast import Module, ImportFrom, Expr, Call, Name, FunctionDef, Assign, Str
from ast import dump, If, Compare, Eq, For, Attribute, Gt, Num, IsNot, BinOp
from ast import NameConstant, Mult, Add, Import, List, Dict, Is, BoolOp, And
from ast import Subscript, Index, Tuple, Lt, Sub, Global, Return, AugAssign
from ast import While, UnaryOp, Not, ClassDef, Mod, Yield, NotEq, Try, Pass
from ast import ExceptHandler, Break, Slice, USub, ListComp, In, Lambda, BitAnd
from ast import BitOr, Or, Delete, Bytes, Raise, NotIn, RShift, GeneratorExp
from ast import Assert, Set, SetComp, LtE, IfExp, FloorDiv, GtE, With, Continue
from ast import YieldFrom, UAdd, LShift, DictComp, Div, Starred, BitXor, Pow
from _ast import arguments, arg as arg_type, keyword as keyword_type
from _ast import alias as alias_type, comprehension, withitem
try:
    # python3.5 support
    from _ast import AsyncFor, AsyncFunctionDef, AsyncWith, Await
except ImportError:
    AsyncFor= AsyncFunctionDef= AsyncWith= Await= object()

class pprint:
    """Lazy pprinter that only does something when converted to string"""
    def __init__ (self, node):
        self.node= node

    def pprint_body (self, body, level):
        for statement in body:
            yield '    '*level
            yield from self.pprint_inner (statement, level)
            yield '\n'

    def pprint_seq (self, seq, sep=', '):
        for index, elem in enumerate (seq):
            if type (elem)==str:
                yield elem
            else:
                yield from self.pprint_inner (elem)

            if index<len (seq)-1:
                if type (sep)==str:
                    yield sep
                else:
                    yield from self.pprint_inner (sep)

    def pprint_orelse (self, orelse, level):
        if len (orelse)>0:
            yield '    '*level+'else:\n'
            yield from self.pprint_body (orelse, level+1)

    def pprint_args (self, args, defaults):
        # TODO: anotations
        # args=[arg(arg='a', annotation=None), arg(arg='b', annotation=None)]
        # defaults=[Num(n=1)]
        d_index= len (args)-len (defaults)
        for index, arg in enumerate (args):
            yield arg.arg

            if index>=d_index:
                yield '='
                yield from self.pprint_inner (defaults[index-d_index])

            if index<len (args)-1:
                yield ', '

    def __str__ (self):
        return ''.join (self.pprint_inner (self.node, 0))

    def pprint_inner (self, node, level=0):
        t= type (node)

        if t==Add:
            yield '+'

        elif t==And:
            yield ' and '

        elif t==Assert:
            # Assert(test=..., msg=None)
            yield 'assert '
            yield from self.pprint_inner (node.test)
            # TODO: msg

        elif t==Assign:
            # Assign(targets=[Name(id='c', ctx=Store())],
            #        value=...)
            yield from self.pprint_seq (node.targets)
            yield '= '
            yield from self.pprint_inner (node.value)

        elif t==AsyncFor:
            yield 'async '
            # For(target=..., iter=..., body=[...], orelse=[...])
            node= For (target=node.target, iter=node.iter, body=node.body, orelse=node.orelse)
            yield from self.pprint_inner (node)

        elif t==AsyncFunctionDef:
            yield 'async '
            # FunctionDef(name='foo', args=arguments(...), body=[ ... ], decorator_list=[], returns=None)
            node= FunctionDef (name=node.name, args=node.args, body=node.body, decorator_list=node.decorator_list,
                               returns=node.returns)
            yield from self.pprint_inner (node)

        elif t==AsyncWith:
            yield 'async '
            # With(items=[...], body=[...])
            node= With (items=node.items, body=node.body)
            yield from self.pprint_inner (node)

        elif t==Attribute:
            # Attribute(value=Name(id='node', ctx=Load()), attr='body', ctx=Load())
            yield from self.pprint_inner (node.value)
            yield '.'
            yield node.attr

        elif t==AugAssign:
            # AugAssign(target=Name(id='ans', ctx=Store()), op=Add(), value=Name(id='a', ctx=Load()))
            yield from self.pprint_inner (node.target)
            yield from self.pprint_inner (node.op)
            yield '= '
            yield from self.pprint_inner (node.value)

        elif t==Await:
            # value=Await(value=...)
            yield 'await '
            yield from self.pprint_inner (node.value)

        elif t==BinOp:
            # BUG: 5*(3+4) -> '5*3+4'
            # do not surround by parenthesis
            # because it breaks part of the code that (ab) uses this to rebuild
            # option and executable names
            # yield '('
            yield from self.pprint_inner (node.left)
            yield from self.pprint_inner (node.op)
            yield from self.pprint_inner (node.right)
            # yield ')'

        elif t==BitAnd:
            yield ' & '

        elif t==BitOr:
            yield '|'

        elif t==BitXor:
            yield '^'

        elif t==BoolOp:
            self.pprint_seq (node.values, node.op)

        elif t==Break:
            yield 'break'

        elif t==Bytes:
            yield repr (node.s)

        elif t==Call:
            # Call(func=Name(id='foo', ctx=Load()), args=[], keywords=[], starargs=None, kwargs=None)
            # TODO: annotations
            yield from self.pprint_inner (node.func)
            yield ' ('
            yield from self.pprint_seq (node.args)

            if len (node.args)>0 and (len (node.keywords)>0 or
                                      node.starargs is not None or
                                      node.kwargs is not None):
                yield ', '

            yield from self.pprint_seq (node.keywords)

            if ((len (node.args)>0 or len (node.keywords)>0) and
                (node.starargs is not None or node.kwargs is not None)):
                yield ', '

            if node.starargs is not None:
                yield '*'
                yield from self.pprint_inner (node.starargs)

            if ((len (node.args)>0 or
                 len (node.keywords)>0 or
                 (node.starargs is not None) and node.kwargs is not None)):
                yield ', '

            if node.kwargs is not None:
                yield '**'
                yield from self.pprint_inner (node.kwargs)

            yield ')'

        elif t==ClassDef:
            # ClassDef(name='ToExpand', bases=[Name(id='object', ctx=Load())],
            #          keywords=[], starargs=None, kwargs=None, body=[...]
            yield 'class '
            yield node.name

            # TODO: more
            if len (node.bases)>0:
                yield ' ('
                yield from self.pprint_seq (node.bases)
                yield ')'

            yield ':'
            yield from self.pprint_body (node.body, level+1)

        elif t==Compare:
            # Compare(left=Name(id='t', ctx=Load()), ops=[Eq()], comparators=[Name(id='Module', ctx=Load())])
            # TODO: do properly
            yield from self.pprint_inner (node.left)
            for op in node.ops:
                yield from self.pprint_inner (op)

            for comparator in node.comparators:
                yield from self.pprint_inner (comparator)

        elif t==Continue:
            yield 'continue'

        elif t==Delete:
            yield 'delete '
            yield from self.pprint_seq (node.targets)

        elif t==Dict:
            yield '{ '
            for k, v in zip (node.keys, node.values):
                yield from self.pprint_inner (k)
                yield '='
                yield from self.pprint_inner (v)
                yield ', '
            yield ' }'

        elif t==DictComp:
            # DictComp(key=Name(id='v', ctx=Load()), value=Name(id='k', ctx=Load()), generators=[comprehension(target=Tuple(elts=[Name(id='k', ctx=Store()), Name(id='v', ctx=Store())], ctx=Store()), iter=Call(func=Name(id='enumerate', ctx=Load()), args=[Name(id='_b32alphabet', ctx=Load())], keywords=[], starargs=None, kwargs=None), ifs=[])])
            yield '{ '
            yield from self.pprint_inner (node.key)
            yield ': '
            yield from self.pprint_inner (node.value)
            yield ' for '
            # TODO: more
            yield from self.pprint_inner (node.generators[0])
            yield ' }'

        elif t==Div:
            yield '/'

        elif t==Eq:
            yield '=='

        elif t==ExceptHandler:
            # ExceptHandler(type=Name(id='KeyError', ctx=Load()), name=None, body=[Pass()])
            yield '    '*level+'except '
            if node.type is not None:
                yield from self.pprint_inner (node.type)
                if node.name is not None:
                    yield ' as '
                    yield node.name

            yield ':\n'
            yield from self.pprint_body (node.body, level+1)

        elif t==Expr:
            # Expr(value=...)
            yield from self.pprint_inner (node.value, level)

        elif t==FloorDiv:
            yield '\\\\'

        elif t==For:
            # For(target=..., iter=..., body=[...], orelse=[...])
            yield 'for '
            yield from self.pprint_inner (node.target)
            yield ' in '
            yield from self.pprint_inner (node.iter)
            yield ':\n'
            yield from self.pprint_body (node.body, level+1)
            yield from self.pprint_orelse (node.orelse, level)

        elif t==FunctionDef:
            # FunctionDef(name='foo', args=arguments(...), body=[ ... ], decorator_list=[], returns=None)
            # TODO: decorator_list
            # TODO: returns
            yield 'def '
            yield node.name
            yield ' ('
            yield from self.pprint_inner (node.args)
            yield '):\n'
            yield from self.pprint_body (node.body, level+1)

        elif t==GeneratorExp:
            # GeneratorExp(elt=Name(id='line', ctx=Load()), generators=[...])
            yield '( '
            yield from self.pprint_inner (node.elt)
            yield ' for '
            # TODO: more
            yield from self.pprint_inner (node.generators[0])
            yield ' )'

        elif t==Global:
            yield 'global '
            yield from self.pprint_seq (node.names)

        elif t==Gt:
            yield '>'

        elif t==GtE:
            yield '>='

        elif t==If:
            # If(test=..., body=[...], orelse=[...]
            yield 'if '
            yield from self.pprint_inner (node.test)
            yield ':\n'
            yield from self.pprint_body (node.body, level+1)

            if len (node.orelse)>0:
                # special case for elif
                if len (node.orelse)==1 and type (node.orelse[0])==If:
                    yield '    '*level+'el'
                    yield from self.pprint_inner (node.orelse[0], level)
                else:
                    yield from self.pprint_orelse (node.orelse, level)

        elif t==IfExp:
            # IfExp(test=..., body=Str(s=''), orelse=Str(s='s'))
            yield from self.pprint_inner (node.body)
            yield ' if '
            yield from self.pprint_inner (node.test)
            yield ' else '
            yield from self.pprint_inner (node.orelse)

        elif t==Import:
            # Import(names=[alias(name='ayrton', asname=None)])
            yield "import "
            yield from self.pprint_seq (node.names)

        elif t==ImportFrom:
            # ImportFrom(module='ayrton.execute', names=[alias(name='Command', asname=None)], level=0)
            yield "from "
            yield node.module
            yield " import "
            yield from self.pprint_seq (node.names)

        elif t==In:
            yield ' in '

        elif t==Index:
            yield from self.pprint_inner (node.value)

        elif t==Is:
            yield ' is '

        elif t==IsNot:
            yield ' is not '

        elif t==LShift:
            yield '<<'

        elif t==Lambda:
            # Lambda(args=arguments(args=[], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=Num(n=0))
            yield 'lambda '
            yield from self.pprint_inner (node.args)
            yield ': '
            yield from self.pprint_inner (node.body)

        elif t==List:
            yield '[ '
            yield from self.pprint_seq (node.elts)
            yield ' ]'

        elif t==ListComp:
            # ListComp(elt=Name(id='i', ctx=Load()), generators=[...])
            # [ i for i in self.indexes if i.right is not None ]
            yield '[ '
            yield from self.pprint_inner (node.elt)
            yield ' for '
            # TODO: more
            yield from self.pprint_inner (node.generators[0])
            yield ' ]'

        elif t==Lt:
            yield '<'

        elif t==LtE:
            yield '<='

        elif t==Mod:
            yield ' % '

        elif t==Module:
            # Module(body=[ ... ])
            yield from self.pprint_body (node.body, 0)

        elif t==Mult:
            yield '*'

        elif t==Name:
            yield node.id

        elif t==NameConstant:
            yield node.value

        elif t==Not:
            yield 'not '

        elif t==NotEq:
            yield '!='

        elif t==NotIn:
            yield ' not in '

        elif t==Num:
            yield str (node.n)

        elif t==Or:
            yield ' or '

        elif t==Pass:
            yield 'pass\n'

        elif t==Pow:
            yield '**'

        elif t==RShift:
            yield '>>'

        elif t==Raise:
            # Raise(exc=Call(func=Name(id='ValueError', ctx=Load()),
            #                args=[Str(s='too many lines')], keywords=[],
            #                starargs=None, kwargs=None),
            #       cause=None)
            yield 'raise '
            if node.exc is not None:
                yield from self.pprint_inner (node.exc)
            # TODO: cause?

        elif t==Return:
            yield 'return '
            if node.value is not None:
                yield from self.pprint_inner (node.value)

        elif t==Set:
            yield '{ '
            self.pprint_seq (node.elts)
            yield ' }'

        elif t==SetComp:
            # SetComp(elt=Name(id='name', ctx=Load()), generators=[...])
            yield '{ '
            yield from self.pprint_inner (node.elt)
            yield ' for '
            # TODO: more
            yield from self.pprint_inner (node.generators[0])

        elif t==Slice:
            # Slice(lower=None, upper=Name(id='left_cb', ctx=Load()), step=None)
            if node.lower is not None:
                yield from self.pprint_inner (node.lower)

            yield ':'

            if node.upper is not None:
                yield from self.pprint_inner (node.upper)

            if node.step is not None:
                yield ':'
                yield from self.pprint_inner (node.step)

        elif t==Str:
            # Str(s='true')
            yield repr (node.s)

        elif t==Sub:
            yield '-'

        elif t==Subscript:
            # Subscript(value=Attribute(value=Name(id='node', ctx=Load()), attr='orelse', ctx=Load()),
            #           slice=Index(value=Num(n=0)), ctx=Load())
            yield from self.pprint_inner (node.value)
            yield '['
            yield from self.pprint_inner (node.slice)
            yield ']'

        elif t==Try:
            # Try(body=[...],  handlers=[...], orelse=[], finalbody=[])
            yield 'try:\n'
            yield from self.pprint_body (node.body, level+1)
            if len (node.handlers)>0:
                for handler in node.handlers:
                    yield from self.pprint_inner (handler, level)

            yield from self.pprint_orelse (node.orelse, level)
            if len (node.finalbody)>0:
                yield '    '*level+'finally:\n'
                yield from self.pprint_body (node.finalbody, level+1)

        elif t==Tuple:
            yield '( '
            yield from self.pprint_seq (node.elts)
            yield ' )'

        elif t==UAdd:
            yield '+'

        elif t==USub:
            yield '-'

        elif t==UnaryOp:
            yield from self.pprint_inner (node.op)
            yield from self.pprint_inner (node.operand)

        elif t==While:
            yield 'while '
            yield from self.pprint_inner (node.test)
            yield ':\n'
            yield from self.pprint_body (node.body, level+1)
            yield from self.pprint_orelse (node.orelse, level)

        elif t==With:
            yield 'with '
            yield from self.pprint_seq (node.items)
            yield ':\n'
            yield from self.pprint_body (node.body, level+1)

        elif t==Yield:
            # Yield(value=Attribute(value=Name(id='self', ctx=Load()), attr='left', ctx=Load()))
            yield 'yield '
            yield from self.pprint_inner (node.value)

        elif t==YieldFrom:
            yield 'yield from '
            yield from self.pprint_inner (node.value)

        elif t==alias_type:
            yield node.name
            if node.asname is not None:
                yield " as "
                yield node.asname

        elif t==arg_type:
            # arg(arg='node', annotation=None)
            # TODO: annotation
            yield node.arg

        elif t==arguments:
            # arguments(args=[arg(arg='a', annotation=None), arg(arg='b', annotation=None)],
            #           vararg=arg(arg='more', annotation=None), kwonlyargs=[], kw_defaults=[],
            #           kwarg=arg(arg='kmore', annotation=None), defaults=[Num(n=1)])

            # this is tricky

            # first there are five, not four, types of arguments
            # positional, positional with default value, extra positional, keywords
            # and extra keywords

            # positional arguments are in args, and the default values in defaults
            # but you have to calculate to which args they belong

            # extra positional is in vararg

            # keyword arguments are in kwonlyargs and the defaults in kw_defaults

            # extra keywords is in kwarg

            yield from self.pprint_args (node.args, node.defaults)

            if len (node.args)>0 and (node.vararg is not None or
                                      len (node.kwonlyargs)>0 or
                                      node.kwarg is not None):
                yield ', '

            if node.vararg is not None:
                yield '*'
                yield from self.pprint_inner (node.vararg)

            if ((len (node.args)>0 or node.vararg is not None) and
                (len (node.kwonlyargs)>0 or node.kwarg is not None)):

                yield ', '

            yield from self.pprint_args (node.kwonlyargs, node.kw_defaults)

            if ((len (node.args)>0 or
                 node.vararg is not None or
                 len (node.kwonlyargs)>0) and node.kwarg is not None):
                yield ', '

            # empty arguments() (from a lambda) do not have this attr (!!!)
            if hasattr (node, 'kwarg') and node.kwarg is not None:
                yield '**'
                yield from self.pprint_inner (node.kwarg)

        elif t==comprehension:
            # comprehension(target=Name(id='i', ctx=Store()),
            #               iter=Attribute(value=Name(id='self', ctx=Load()),
            #                              attr='indexes', ctx=Load()),
            #               ifs=[Compare(left=..., ops=[IsNot()],
            #                            comparators=[NameConstant(value=None)])])
            # i in self.indexes if i.right is not None
            yield from self.pprint_inner (node.target)
            yield ' in '
            yield from self.pprint_inner (node.iter)

            if len (node.ifs)>0:
                # TODO: more
                yield ' if '
                yield from self.pprint_inner (node.ifs[0])

        elif t==keyword_type:
            # keyword(arg='end', value=Str(s=''))
            yield node.arg
            yield '='
            yield from self.pprint_inner (node.value)

        elif t==withitem:
            # withitem(context_expr=..., optional_vars=Name(id='f', ctx=Store()))
            yield from self.pprint_inner (node.context_expr)
            if node.optional_vars is not None:
                yield ' as '
                yield from self.pprint_inner (node.optional_vars)

        elif t==str:
            yield node

        else:
            yield '\n'
            yield '# unknown construction\n'
            yield dump (node)



import io
import itertools
import os
import sys
import traceback
from errno import EINTR

from ayrton.parser import debug


AUTO_DEBUG = os.getenv('PYPY_DEBUG')
RECORD_INTERPLEVEL_TRACEBACK = True

def strerror(errno):
    """Translate an error code to a message string."""
    from pypy.module._codecs.locale import str_decode_locale_surrogateescape
    return str_decode_locale_surrogateescape(os.strerror(errno))

class OperationError(Exception):
    """Interpreter-level exception that signals an exception that should be
    sent to the application level.

    OperationError instances have three attributes (and no .args),
    w_type, _w_value and _application_traceback, which contain the wrapped
    type and value describing the exception, and a chained list of
    PyTraceback objects making the application-level traceback.
    """

    _w_value = None
    _application_traceback = None
    w_cause = None

    def __init__(self, w_type, w_value, tb=None, w_cause=None):
        self.setup(w_type, w_value)
        self._application_traceback = tb
        self.w_cause = w_cause

    def setup(self, w_type, w_value=None):
        assert w_type is not None
        self.w_type = w_type
        self._w_value = w_value
        if not we_are_translated():
            self.debug_excs = []

    def clear(self, space):
        # XXX remove this method.  The point is that we cannot always
        # hack at 'self' to clear w_type and _w_value, because in some
        # corner cases the OperationError will be used again: see
        # test_interpreter.py:test_with_statement_and_sys_clear.
        pass

    def match(self, space, w_check_class):
        "Check if this application-level exception matches 'w_check_class'."
        return space.exception_match(self.w_type, w_check_class)

    def async(self, space):
        "Check if this is an exception that should better not be caught."
        return (self.match(space, space.w_SystemExit) or
                self.match(space, space.w_KeyboardInterrupt))

    def __str__(self):
        "NOT_RPYTHON: Convenience for tracebacks."
        s = self._w_value
        if self.__class__ is not OperationError and s is None:
            space = getattr(self.w_type, 'space')
            if space is not None:
                s = self._compute_value(space)
        return '[%s: %s]' % (self.w_type, s)

    def __repr__(self):
        "NOT_RPYTHON"
        return 'OperationError(%s)' % (self.w_type)

    def errorstr(self, space, use_repr=False):
        "The exception class and value, as a string."
        w_value = self.get_w_value(space)
        if space is None:
            # this part NOT_RPYTHON
            exc_typename = str(self.w_type)
            exc_value = str(w_value)
        else:
            w = space.wrap
            if space.is_w(space.type(self.w_type), space.w_text):
                exc_typename = space.str_w(self.w_type)
            else:
                exc_typename = space.str_w(
                    space.getattr(self.w_type, w('__name__')))
            if space.is_w(w_value, space.w_None):
                exc_value = ""
            else:
                try:
                    if use_repr:
                        exc_value = space.str_w(space.repr(w_value))
                    else:
                        exc_value = space.str_w(space.str(w_value))
                except OperationError:
                    # oups, cannot __str__ the exception object
                    exc_value = "<oups, exception object itself cannot be str'd>"
        if not exc_value:
            return exc_typename
        else:
            return '%s: %s' % (exc_typename, exc_value)

    def record_interpreter_traceback(self):
        """Records the current traceback inside the interpreter.
        This traceback is only useful to debug the interpreter, not the
        application."""
        if not we_are_translated():
            if RECORD_INTERPLEVEL_TRACEBACK:
                self.debug_excs.append(sys.exc_info())

    def print_application_traceback(self, space, file=None):
        "NOT_RPYTHON: Dump a standard application-level traceback."
        if file is None:
            file = sys.stderr
        self.print_app_tb_only(file)
        print >> file, self.errorstr(space)

    def print_app_tb_only(self, file):
        "NOT_RPYTHON"
        tb = self._application_traceback
        if tb:
            import linecache
            print >> file, "Traceback (application-level):"
            while tb is not None:
                co = tb.frame.pycode
                lineno = tb.get_lineno()
                fname = co.co_filename
                if fname.startswith('<inline>\n'):
                    lines = fname.split('\n')
                    fname = lines[0].strip()
                    try:
                        l = lines[lineno]
                    except IndexError:
                        l = ''
                else:
                    l = linecache.getline(fname, lineno)
                print >> file, "  File \"%s\"," % fname,
                print >> file, "line", lineno, "in", co.co_name
                if l:
                    if l.endswith('\n'):
                        l = l[:-1]
                    l = "    " + l.lstrip()
                    print >> file, l
                tb = tb.next

    def print_detailed_traceback(self, space=None, file=None):
        """NOT_RPYTHON: Dump a nice detailed interpreter- and
        application-level traceback, useful to debug the interpreter."""
        if file is None:
            file = sys.stderr
        f = io.StringIO()
        for i in range(len(self.debug_excs)-1, -1, -1):
            print >> f, "Traceback (interpreter-level):"
            traceback.print_tb(self.debug_excs[i][2], file=f)
        f.seek(0)
        debug_print(''.join(['|| ' + line for line in f.readlines()]), file)
        if self.debug_excs:
            from pypy.tool import tb_server
            tb_server.publish_exc(self.debug_excs[-1])
        self.print_app_tb_only(file)
        print >> file, '(application-level)', self.errorstr(space)
        if AUTO_DEBUG:
            debug.fire(self)

    def normalize_exception(self, space):
        """Normalize the OperationError.  In other words, fix w_type and/or
        w_value to make sure that the __class__ of w_value is exactly w_type.
        """
        #
        # This method covers all ways in which the Python statement
        # "raise X, Y" can produce a valid exception type and instance.
        #
        # In the following table, 'Class' means a subclass of BaseException
        # and 'inst' is an instance of either 'Class' or a subclass of it.
        #
        # The flow object space only deals with non-advanced case.
        #
        #  input (w_type, w_value)... becomes...                advanced case?
        # ---------------------------------------------------------------------
        #  (Class, None)              (Class, Class())                no
        #  (Class, inst)              (inst.__class__, inst)          no
        #  (Class, tuple)             (Class, Class(*tuple))          yes
        #  (Class, x)                 (Class, Class(x))               no
        #  (inst, None)               (inst.__class__, inst)          no
        #
        w_type = self.w_type
        w_value = self.get_w_value(space)

        if space.exception_is_valid_obj_as_class_w(w_type):
            # this is for all cases of the form (Class, something)
            if space.is_w(w_value, space.w_None):
                # raise Type: we assume we have to instantiate Type
                w_value = space.call_function(w_type)
                w_type = self._exception_getclass(space, w_value)
            else:
                w_valuetype = space.exception_getclass(w_value)
                if space.exception_issubclass_w(w_valuetype, w_type):
                    # raise Type, Instance: let etype be the exact type of value
                    w_type = w_valuetype
                else:
                    if space.isinstance_w(w_value, space.w_tuple):
                        # raise Type, tuple: assume the tuple contains the
                        #                    constructor args
                        w_value = space.call(w_type, w_value)
                    else:
                        # raise Type, X: assume X is the constructor argument
                        w_value = space.call_function(w_type, w_value)
                    w_type = self._exception_getclass(space, w_value)
            if self.w_cause:
                # ensure w_cause is of a valid type
                if space.is_none(self.w_cause):
                    pass
                else:
                    self._exception_getclass(space, self.w_cause, "exception causes")
                space.setattr(w_value, space.wrap("__cause__"), self.w_cause)
            if self._application_traceback:
                from pypy.interpreter.pytraceback import PyTraceback
                from pypy.module.exceptions.interp_exceptions import W_BaseException
                tb = self._application_traceback
                if (isinstance(w_value, W_BaseException) and
                    isinstance(tb, PyTraceback)):
                    # traceback hasn't escaped yet
                    w_value.w_traceback = tb
                else:
                    # traceback has escaped
                    space.setattr(w_value, space.wrap("__traceback__"),
                                  space.wrap(self.get_traceback()))
        else:
            # the only case left here is (inst, None), from a 'raise inst'.
            w_inst = w_type
            w_instclass = self._exception_getclass(space, w_inst)
            if not space.is_w(w_value, space.w_None):
                raise OperationError(space.w_TypeError,
                                     space.wrap("instance exception may not "
                                                "have a separate value"))
            w_value = w_inst
            w_type = w_instclass

        self.w_type = w_type
        self._w_value = w_value

    def _exception_getclass(self, space, w_inst, what="exceptions"):
        w_type = space.exception_getclass(w_inst)
        if not space.exception_is_valid_class_w(w_type):
            raise oefmt(space.w_TypeError,
                        "%s must derive from BaseException, not %N",
                        what, w_type)
        return w_type

    def write_unraisable(self, space, where, w_object=None,
                         with_traceback=False, extra_line=''):
        if w_object is None:
            objrepr = ''
        else:
            try:
                objrepr = space.str_w(space.repr(w_object))
            except OperationError:
                objrepr = '?'
        #
        try:
            if with_traceback:
                try:
                    self.normalize_exception(space)
                except OperationError:
                    pass
                w_t = self.w_type
                w_v = self.get_w_value(space)
                w_tb = space.wrap(self.get_traceback())
                space.appexec([space.wrap(where),
                               space.wrap(objrepr),
                               space.wrap(extra_line),
                               w_t, w_v, w_tb],
                """(where, objrepr, extra_line, t, v, tb):
                    import sys, traceback
                    if where or objrepr:
                        sys.stderr.write('From %s%s:\\n' % (where, objrepr))
                    if extra_line:
                        sys.stderr.write(extra_line)
                    traceback.print_exception(t, v, tb)
                """)
            else:
                msg = 'Exception %s in %s%s ignored\n' % (
                    self.errorstr(space, use_repr=True), where, objrepr)
                space.call_method(space.sys.get('stderr'), 'write',
                                  space.wrap(msg))
        except OperationError:
            pass   # ignored

    def get_w_value(self, space):
        w_value = self._w_value
        if w_value is None:
            value = self._compute_value(space)
            self._w_value = w_value = space.wrap(value)
        return w_value

    def _compute_value(self, space):
        raise NotImplementedError

    def get_traceback(self):
        """Calling this marks the PyTraceback as escaped, i.e. it becomes
        accessible and inspectable by app-level Python code.  For the JIT.
        Note that this has no effect if there are already several traceback
        frames recorded, because in this case they are already marked as
        escaping by executioncontext.leave() being called with
        got_exception=True.
        """
        from pypy.interpreter.pytraceback import PyTraceback
        tb = self._application_traceback
        if tb is not None and isinstance(tb, PyTraceback):
            tb.frame.mark_as_escaped()
        return tb

    def set_traceback(self, traceback):
        """Set the current traceback.  It should either be a traceback
        pointing to some already-escaped frame, or a traceback for the
        current frame.  To support the latter case we do not mark the
        frame as escaped.  The idea is that it will be marked as escaping
        only if the exception really propagates out of this frame, by
        executioncontext.leave() being called with got_exception=True.
        """
        self._application_traceback = traceback

    def remove_traceback_module_frames(self, module_name):
        from pypy.interpreter.pytraceback import PyTraceback
        tb = self._application_traceback
        while tb is not None and isinstance(tb, PyTraceback):
            if tb.frame.pycode.co_filename != module_name:
                break
            tb = tb.next
        self._application_traceback = tb

    def record_context(self, space, frame):
        """Record a __context__ for this exception from the current
        frame if one exists.

        __context__ is otherwise lazily determined from the
        traceback. However the current frame.last_exception must be
        checked for a __context__ before this OperationError overwrites
        it (making the previous last_exception unavailable later on).
        """
        last_exception = frame.last_exception
        if (last_exception is not None and not frame.hide() or
            last_exception is get_cleared_operation_error(space)):
            # normalize w_value so setup_context can check for cycles
            self.normalize_exception(space)
            w_value = self.get_w_value(space)
            w_last = last_exception.get_w_value(space)
            w_context = setup_context(space, w_value, w_last, lazy=True)
            space.setattr(w_value, space.wrap('__context__'), w_context)


def oefmt(w_type, valuefmt, *args):
    """Equivalent to OperationError(w_type, space.wrap(valuefmt % args)).
    More efficient in the (common) case where the value is not actually
    needed. Note that in the py3k branch the exception message will
    always be unicode.

    Supports the standard %s and %d formats, plus the following:

    %8 - The result of arg.decode('utf-8')
    %N - The result of w_arg.getname(space)
    %R - The result of space.unicode_w(space.repr(w_arg))
    %T - The result of space.type(w_arg).name

    """
    if not len(args):
        return OpErrFmtNoArgs(w_type, valuefmt)
    OpErrFmt, strings = get_operr_class(valuefmt)
    return OpErrFmt(w_type, strings, *args)



# ______________________________________________________________________
"""Module automata

THIS FILE WAS COPIED FROM pypy/module/parser/pytokenize.py AND ADAPTED
TO BE ANNOTABLE (Mainly made the DFA's __init__ accept two lists
instead of a unique nested one)

$Id: automata.py,v 1.2 2003/10/02 17:37:17 jriehl Exp $
"""
# ______________________________________________________________________
# Module level definitions

# PYPY Modification: removed the EMPTY class as it's not needed here


# PYPY Modification: we don't need a particuliar DEFAULT class here
#                    a simple None works fine.
#                    (Having a DefaultClass inheriting from str makes
#                     the annotator crash)
DEFAULT = "\00default" # XXX hack, the rtyper does not support dict of with str|None keys
                       # anyway using dicts doesn't seem the best final way to store these char indexed tables
# PYPY Modification : removed all automata functions (any, maybe,
#                     newArcPair, etc.)

class DFA:
    # ____________________________________________________________
    def __init__(self, states, accepts, start = 0):
        self.states = states
        self.accepts = accepts
        self.start = start

    # ____________________________________________________________
    def recognize (self, inVec, pos = 0): # greedy = True
        crntState = self.start
        lastAccept = False
        i = pos
        for i in range(pos, len(inVec)):
            item = inVec[i]
            if ord(item) > 0x80:
                item = "\x80"  # NON_ASCII
            # arcMap, accept = self.states[crntState]
            arcMap = self.states[crntState]
            accept = self.accepts[crntState]
            if item in arcMap:
                crntState = arcMap[item]
            elif DEFAULT in arcMap:
                crntState = arcMap[DEFAULT]
            elif accept:
                return i
            elif lastAccept:
                # This is now needed b/c of exception cases where there are
                # transitions to dead states
                return i - 1
            else:
                return -1
            lastAccept = accept
        # if self.states[crntState][1]:
        if self.accepts[crntState]:
            return i + 1
        elif lastAccept:
            return i
        else:
            return -1

# ______________________________________________________________________

class NonGreedyDFA (DFA):
    def recognize (self, inVec, pos = 0):
        crntState = self.start
        i = pos
        for item in inVec[pos:]:
            # arcMap, accept = self.states[crntState]
            arcMap = self.states[crntState]
            accept = self.accepts[crntState]
            if accept:
                return i
            elif item in arcMap:
                crntState = arcMap[item]
            elif DEFAULT in arcMap:
                crntState = arcMap[DEFAULT]
            else:
                return -1
            i += 1
        # if self.states[crntState][1]:
        if self.accepts[crntState]:
            return i
        else:
            return -1

# ______________________________________________________________________
# End of automata.py



# -*- coding: utf-8 -*-

# (c) 2013 Marcos Dione <mdione@grulic.org.ar>

# This file is part of ayrton.
#
# ayrton is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ayrton is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ayrton.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import importlib
import ast
import logging
import dis
import traceback

# patch logging so we have debug2 and debug3
import ayrton.utils

log_format= "%(asctime)s %(name)16s:%(lineno)-4d (%(funcName)-21s) %(levelname)-8s %(message)s"
date_format= "%H:%M:%S"

# uncomment one of these for way too much debugging :)
# logging.basicConfig(filename='ayrton.%d.log' % os.getpid (), level=logging.DEBUG, format=log_format, datefmt=date_format)
# logging.basicConfig(filename='ayrton.log', level=logging.DEBUG, format=log_format, datefmt=date_format)
logger= logging.getLogger ('ayrton')

# things that have to be defined before importing ayton.execute :(
# singleton needed so the functions can access the runner
runner= None

from ayrton.castt import CrazyASTTransformer
from ayrton.execute import o, Command, Capture, CommandFailed
from ayrton.parser.pyparser.pyparse import CompileInfo, PythonParser
from ayrton.parser.astcompiler.astbuilder import ast_from_node
from ayrton.ast_pprinter import pprint

__version__= '0.7'


def parse (script, file_name=''):
    parser= PythonParser (None)
    info= CompileInfo (file_name, 'exec')
    return ast_from_node (None, parser.parse_source (script, info), info)


class Environment (dict):
    def __init__ (self, *args, **kwargs):
        super ().__init__ (*args, **kwargs)

        self.polute ()


    def polute (self):
        self.update (__builtins__)
        # weed out some stuff
        for weed in ('copyright', '__doc__', 'help', '__package__', 'credits',
                     'license', '__name__', 'quit', 'exit'):
            del self[weed]

        # these functions will be loaded from each module and put in the globals
        # tuples (src, dst) renames function src to dst
        ayrton_builtins= {
            'os': [ ('getcwd', 'pwd'), 'uname', 'listdir', ],
            'os.path': [ 'abspath', 'basename', 'commonprefix', 'dirname',  ],
            'time': [ 'sleep', ],
            'sys': [ 'exit', ],  # argv is handled just before execution

            'ayrton.file_test': [ '_a', '_b', '_c', '_d', '_e', '_f', '_g', '_h',
                                  '_k', '_p', '_r', '_s', '_u', '_w', '_x', '_L',
                                  '_N', '_S', '_nt', '_ot' ],
            'ayrton.expansion': [ 'bash', ],
            'ayrton.functions': [ 'cd', ('cd', 'chdir'), 'export', 'option', 'run',
                                   'shift', 'unset', ],
            'ayrton.execute': [ 'o', 'Capture', 'CommandFailed', 'CommandNotFound',
                                'Pipe', 'Command'],
            'ayrton.remote': [ 'remote' ]
            }

        for module, functions in ayrton_builtins.items ():
            m= importlib.import_module (module)
            for function in functions:
                if type (function)==tuple:
                    src, dst= function
                else:
                    src= function
                    dst= function

                self[dst]= getattr (m, src)

        # now the IO files
        for std in ('stdin', 'stdout', 'stderr'):
            self[std]= getattr (sys, std).buffer


class Ayrton (object):
    def __init__ (self, g=None, l=None, **kwargs):
        logger.debug ('===========================================================')
        logger.debug ('new interpreter')
        logger.debug3 ('globals: %s', ayrton.utils.dump_dict (g))
        logger.debug3 ('locals: %s', ayrton.utils.dump_dict (l))

        self.globals= Environment ()
        if g is not None:
            self.globals.update (g)
        self.globals.update (kwargs)

        if l is None:
            # If exec gets two separate objects as globals and locals,
            # the code will be executed as if it were embedded in a class definition.
            # and this happens:
            """
            In [7]: source='''import math
            ...: def foo ():
            ...:     math.floor (1.1, )
            ...:
            ...: foo()'''

            In [8]: import ast
            In [9]: t= ast.parse (source)
            In [10]: c= compile (t, 'foo.py', 'exec')
            In [11]: exec (c, {}, {})
            ---------------------------------------------------------------------------
            NameError                                 Traceback (most recent call last)
            <ipython-input-11-3a2844c232c1> in <module>()
            ----> 1 exec (c, {}, {})
            /home/mdione/src/projects/ayrton/foo.py in <module>()
            /home/mdione/src/projects/ayrton/foo.py in foo()
            NameError: name 'math' is not defined
            """
            self.locals= self.globals
        else:
            self.locals= l

        self.options= {}
        self.pending_children= []

        # HACK to update the singleton
        # this might break if we implement subinstances
        global runner
        runner= self


    def run_file (self, file_name, argv=None):
        # it's a pity that parse() does not accept a file as input
        # so we could avoid reading the whole file
        logger.debug ('running from file %s', file_name)

        f= open (file_name)
        script= f.read ()
        f.close ()

        return self.run_script (script, file_name, argv)


    def run_script (self, script, file_name, argv=None):
        logger.debug ('running script:\n-----------\n%s\n-----------', script)
        tree= parse (script, file_name)
        # TODO: self.locals?
        tree= CrazyASTTransformer (self.globals, file_name).modify (tree)

        return self.run_tree (tree, file_name, argv)


    def run_tree (self, tree, file_name, argv=None):
        logger.debug2 ('AST: %s', ast.dump (tree))
        logger.debug2 ('code: \n%s', pprint (tree))

        return self.run_code (compile (tree, file_name, 'exec'), file_name, argv)


    def run_code (self, code, file_name, argv=None):
        if logger.parent.level<=logging.DEBUG2:
            logger.debug2 ('------------------')
            logger.debug2 ('main (gobal) code:')
            handler= logger.parent.handlers[0]

            handler.acquire ()
            dis.dis (code, file=handler.stream)
            handler.release ()

            for inst in dis.Bytecode (code):
                if inst.opname=='LOAD_CONST':
                    if type (inst.argval)==type (code):
                        logger.debug ('------------------')
                        handler.acquire ()
                        dis.dis (inst.argval, file=handler.stream)
                        handler.release ()
                    elif type (inst.argval)==str:
                        logger.debug ("last function is called: %s", inst.argval)

        # prepare environment
        self.globals.update (os.environ)

        logger.debug (argv)
        if argv is None:
            argv= [ file_name ]
        self.globals['argv']= argv

        '''
        exec(): If only globals is provided, it must be a dictionary, which will
        be used for both the global and the local variables. If globals and locals
        are given, they are used for the global and local variables, respectively.
        If provided, locals can be any mapping object. Remember that at module
        level, globals and locals are the same dictionary. If exec gets two
        separate objects as globals and locals, the code will be executed as if
        it were embedded in a class definition.

        If the globals dictionary does not contain a value for the key __builtins__,
        a reference to the dictionary of the built-in module builtins is inserted
        under that key. That way you can control what builtins are available to
        the executed code by inserting your own __builtins__ dictionary into
        globals before passing it to exec().

        The default locals act as described for function locals() below:
        modifications to the default locals dictionary should not be attempted.
        Pass an explicit locals dictionary if you need to see effects of the code
        on locals after function exec() returns.

        locals(): Update and return a dictionary representing the current local
        symbol table. Free variables are returned by locals() when it is called
        in function blocks, but not in class blocks.

        The contents of this dictionary should not be modified; changes may not
        affect the values of local and free variables used by the interpreter.
        '''
        error= None
        try:
            logger.debug3 ('globals for script: %s', ayrton.utils.dump_dict (self.globals))
            exec (code, self.globals, self.locals)
        except Exception as e:
            logger.debug ('script finished by Exception')
            logger.debug (traceback.format_exc ())
            error= e

        logger.debug3 ('globals at script exit: %s', ayrton.utils.dump_dict (self.globals))
        logger.debug3 ('locals at script exit: %s', ayrton.utils.dump_dict (self.locals))
        result= self.locals.get ('ayrton_return_value', None)
        logger.debug ('ayrton_return_value: %r', result)

        if error is not None:
            raise error

        return result


    def wait_for_pending_children (self):
        for i in range (len (self.pending_children)):
            child= self.pending_children.pop (0)
            child.wait ()


def run_tree (tree, g, l):
    """main entry point for remote()"""
    runner= Ayrton (g=g, l=l)
    return runner.run_tree (tree, 'unknown_tree')

def run_file_or_script (script=None, file_name='script_from_command_line', argv=None,
                        **kwargs):
    """Main entry point for bin/ayrton and unittests."""
    runner= Ayrton (**kwargs)
    if script is None:
        v= runner.run_file (file_name, argv)
    else:
        v= runner.run_script (script, file_name, argv)

    return v

# backwards support for unit tests
main= run_file_or_script



# -*- coding: utf-8 -*-

# (c) 2013 Marcos Dione <mdione@grulic.org.ar>

# This file is part of ayrton.
#
# ayrton is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ayrton is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ayrton.  If not, see <http://www.gnu.org/licenses/>.

from glob import glob
from functools import reduce
from itertools import chain
import os.path

def glob_expand (s):
    # shamelessly insprired from sh.py
    # but we return a list with the string when there is no match
    # and we also handle lists of strings; we expand them independently
    if type (s)==str:
        l= [ s ]
    else:
        # otherwise we assume it's some kind of iterable
        l= s

    ans= []
    for s in l:
        a= glob (s)
        if a==[]:
            a= [ s ]

        # accumulate them
        ans+= a

    return ans

class Group (object):
    def __init__ (self, left, right=None):
        self.left= left
        self.right= right

    def add_right (self, right):
        self.right= right

    def __repr__ (self):
        return "Group (%d, %r)" % (self.left, self.right)

    def __iter__ (self):
        yield self.left
        yield self.right

class ToExpand (object):
    def __init__ (self, s, indexes=None):
        self.text= s
        if indexes is None:
            self.find_indexes ()
        else:
            # I have to deep copy the Group
            self.indexes= []
            for i in indexes:
                self.indexes.append (Group (*list (i)))

    def find_indexes (self):
        self.indexes= []
        current= {}
        seq= 0

        # iterate from left to right over the chars
        for i, c in enumerate (self.text):
            # if the char is {, try to find the first closing }
            # but if another { is found, use that as the matching one for the }
            if   c=='{' and self.text[i-1]!='\\':
                data= Group (i)
                # print (data)
                self.indexes.append (data)
                # we make sure we point to the same list,
                # so we can add the pointer to the closing bracket when we find it
                current[seq]= data
                seq+= 1
            elif c=='}' and self.text[i-1]!='\\':
                # we append to the element the pointer to the closing bracket
                try:
                    data= current[seq-1]
                    data.add_right (i)
                    # print (data)
                    seq-= 1
                except KeyError:
                    # there is no corresponding opening bracket; just ignore
                    pass

        # remove indexes for unmatched open brackes
        # unmatched closing brackets are ignored in the try/except up there
        self.indexes= [ i for i in self.indexes if i.right is not None ]
        # print (self.indexes)

    def expand_one (self):
        "expand the more-to-the-left/outer bracket, return a list of ToExpand's"
        data= self.indexes.pop (0)

        left_cb, right_cb= data
        prefix= self.text[:left_cb]
        # includes the {}'s
        body= self.text[left_cb:right_cb+1]
        postfix= self.text[right_cb+1:]
        # print ("pre:%r b:%r post:%r" % (prefix, body, postfix))

        # will hold the expansion of the body
        expanded= []
        # do not count the opening bracket
        last= 1
        split_here= False
        comma_found= False
        # index for the position in the original string of the beginning of
        # the current body starts
        # we need it because the indexes are base on the positions on the original string
        base= left_cb+1

        for i, c in enumerate (body):
            if c==',' and body[i-1]!='\\':
                split_here= True
                # check if this comma belongs to a inner group
                for l, r in self.indexes:
                    # TODO: show that this cannot actually happen
                    if l>right_cb:
                        # stop searching, this pair is beyond us
                        break

                    orig_i= base+i
                    # print (s, l, orig_i, r)

                    if l<orig_i and orig_i<r:
                        # print 'not splitting, comma @%d(%d) between {%d,%d}' % (orig_i, i, l, r)
                        split_here= False
                        # stop searching too
                        break

                if split_here:
                    comma_found= True
                    dst= body[last:i]
                    te= ToExpand (prefix+dst+postfix)
                    expanded.append (te)
                    # print ('split!', expanded)
                    last= i+1 # point to the next char, not the comma itself

        # only add the last element if a comma was found
        if comma_found:
            # do not count the closing bracket
            dst= body[last:-1]
            te= ToExpand (prefix+dst+postfix)
            expanded.append (te)
            # print ('append', expanded)
        else:
            # otherwise, just leave untouched
            # (except for the index we already removed at the beginning of this method)
            expanded.append (self)

        return expanded

    def fully_expanded (self):
        # just check how many indexes are left
        return len (self.indexes)==0

    def __repr__ (self):
        return "ToExpand (%s, %s)" % (self.text, self.indexes)

def brace_expand (s):
    # NOTE: this function is O(N) in several ways
    if type (s)==str:
        l= [ s ]
    else:
        # otherwise we assume it's some kind of iterable
        l= s

    ans= []

    for s in l:
        todo= [ ToExpand (s) ]

        while (len (todo)>0):
            # print (todo)
            te= todo.pop (0)

            if not te.fully_expanded ():
                new_ones= te.expand_one ()
                for index, new in enumerate (new_ones):
                    if new.fully_expanded ():
                        ans.append (new.text)
                    else:
                        todo.insert (index, new)
            else:
                ans.append (te.text)

    return ans

def backslash_descape (s):
    if type (s)==str:
        ans= s.replace ('\\', '')
    else:
        # otherwise we assume it's some kind of iterable
        ans= [ s1.replace ('\\', '') for s1 in s ]

    return ans

def tilde_expand (s):
    if type (s)==str:
        ans= os.path.expanduser (s)
    else:
        # otherwise we assume it's some kind of iterable
        ans= [ os.path.expanduser (s1) for s1 in s ]

    return ans

def bash (s, single=False):
    data= backslash_descape (glob_expand (tilde_expand (brace_expand (s))))
    if single and len(data)==1:
        data= data[0]
    return data
